/*
  Copyright (C) 2002 Paul Davis

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  $Id: gain_pan_meter.cc,v 1.51 2004/07/01 03:48:16 pauld Exp $
*/

#include <limits.h>

#include <ardour/io.h>
#include <ardour/dB.h>
#include <gtkmmext/utils.h>
#include <gtkmmext/fastmeter.h>
#include <gtkmmext/stop_signal.h>
#include <gtkmmext/barcontroller.h>
#include <midi++/manager.h>
#include <pbd/fastlog.h>

#include "ardour_ui.h"
#include "gain_pan_meter.h"
#include "panner2d.h"
#include "utils.h"
#include "logmeter.h"

#include <ardour/session.h>
#include <ardour/route.h>
#include <ardour/panner.h>

#include "i18n.h"

using namespace ARDOUR;
using namespace Gtkmmext;
using namespace Gtk;
using namespace SigC;

/* XPM */
static const gchar * forwdblarrow_xpm[] = {
"20 11 3 1",
" 	c None",
".	c #000000",
"+	c #FFFFFF",
"              ..    ",
"              .+.   ",
"              .++.  ",
" ..............+++. ",
" .+++++++++++++++++.",
" ...................",
" .+++++++++++++++++.",
" ..............+++. ",
"              .++.  ",
"              .+.   ",
"              ..    "};

/* XPM */
static const gchar * revdblarrow_xpm[] = {
"20 11 3 1",
" 	c None",
".	c #000000",
"+	c #FFFFFF",
"              ..    ",
"              .+.   ",
"              .++.  ",
" ..............+++. ",
" .+++++++++++++++++.",
" ...................",
" .+++++++++++++++++.",
"  .+++..............",
"   .++.             ",
"    .+.             ",
"     ..             "};


GainPanMeter::GainPanMeter (IO& io, Session& s)
	: _io (io),
	  _session (s),
	  gain_slider (0),
	  // 0.781787 is the value needed for gain to be set to 0.
	  gain_adjustment (0.781787, 0.0, 1.0, 0.01, 0.1),
	  gain_display (&gain_adjustment, "MixerStripGainDisplay"),
	  gain_unit_label (_("dbFS")),
	  meter_pre_label (_("pre")),
	  top_table (1, 2),
	  panning_up_arrow (GTK_ARROW_UP, GTK_SHADOW_OUT),
	  panning_down_arrow (GTK_ARROW_DOWN, GTK_SHADOW_OUT),
	  panning_link_button (_("link"))
	
{
	ignore_toggle = false;
	meter_menu = 0;
	pan_menu = 0;
	in_pan_update = false;

	gain_slider = manage (new VSliderController (ARDOUR_UI::instance()->mixer_slider_pix(),
						     &gain_adjustment,
						     & _io.midi_gain_control(),
						     false));

	gain_slider->fader().button_press_event.connect (slot (*this, &GainPanMeter::start_gain_touch));
	gain_slider->fader().button_release_event.connect (slot (*this, &GainPanMeter::end_gain_touch));

	/* XXX should we really bind to the MMC port, or a different port? */

	if (_session.mmc_port()) {
		_io.midi_gain_control().midi_rebind (_session.mmc_port(), 0);
		_io.set_midi_to_gain_function (slider_position_to_gain);
	}

	gain_display.set_print_func (_gain_printer, this);
	set_usize_to_display_given_text (gain_display, "-86.0", 2, 2);

	gain_unit_button.add (gain_unit_label);
	gain_unit_button.set_name ("MixerStripGainUnitButton");
	gain_unit_label.set_name ("MixerStripGainUnitButton");

	top_table.set_col_spacings (2);
	top_table.set_homogeneous (true);
	top_table.attach (gain_unit_button, 0, 1, 0, 1);

	Route* r;

	if ((r = dynamic_cast<Route*> (&_io)) != 0) {
		r->meter_change.connect (slot (*this, &GainPanMeter::meter_changed));
		meter_pre_button.add (meter_pre_label);
		meter_pre_button.set_name ("MixerStripMeterPreButton");
		meter_pre_label.set_name ("MixerStripMeterPreButton");
		meter_pre_button.set_active (r->meter_pre_fader());
		meter_pre_button.toggled.connect (slot (*this, &GainPanMeter::meter_pre_toggled));
		top_table.attach (meter_pre_button, 1, 2, 0, 1);
	}

	gain_display_box.set_spacing (2);
	gain_display_frame.set_shadow_type (GTK_SHADOW_IN);
	gain_display_frame.set_name ("BaseFrame");
	gain_display_frame.add (gain_display);
	gain_display_box.pack_start (gain_display_frame, false, false);

	peak_display.set_name ("MixerStripPeakDisplay");
	set_usize_to_display_given_text (peak_display, "-86.0", 2, 2);
	peak_display.add (peak_display_label);
	peak_display_frame.set_shadow_type (GTK_SHADOW_IN);
	peak_display_frame.set_name ("BaseFrame");
	peak_display_frame.add (peak_display);
	max_peak = minus_infinity();
	peak_display_label.set_text (_("-inf"));
	
	gain_display_box.pack_start (peak_display_frame, false, false);

	pan_bar_packer.set_usize (-1, 61);
	panning_viewport.set_usize (61, 61);

	panning_viewport.set_name (X_("BaseFrame"));

	ARDOUR_UI::instance()->tooltips().set_tip (panning_link_button,
						   _("panning link control"));
	ARDOUR_UI::instance()->tooltips().set_tip (panning_link_direction_button,
						   _("panning link direction"));

	panning_link_box.pack_start (panning_link_button, true, true);
	panning_link_box.pack_start (panning_link_direction_button, true, true);

	panning_link_button.set_name (X_("PanningLinkButton"));
	panning_link_direction_button.set_name (X_("PanningLinkDirectionButton"));

	/* the pixmap will be reset at some point, but the key thing is that
	   we need a pixmap in the button just to get started.
	*/

	panning_link_direction_button.add (*(manage (new Pixmap (forwdblarrow_xpm))));

	panning_link_direction_button.clicked.connect
		(slot (*this, &GainPanMeter::panning_link_direction_clicked));

	panning_link_button.button_press_event.connect
		(slot (*this, &GainPanMeter::panning_link_button_press));
	panning_link_button.button_release_event.connect
		(slot (*this, &GainPanMeter::panning_link_button_release));

	panning_up.set_border_width (3);
	panning_down.set_border_width (3);
	panning_up.add (panning_up_arrow);
	panning_down.add (panning_down_arrow);
	panning_up.set_name (X_("PanScrollerBase"));
	panning_down.set_name (X_("PanScrollerBase"));
	panning_up_arrow.set_name (X_("PanScrollerArrow"));
	panning_down_arrow.set_name (X_("PanScrollerArrow"));

	panning_scroll_box.pack_start (panning_up, true, true);
	panning_scroll_box.pack_start (panning_down, true, true);

	pan_vbox.set_spacing (4);
	// pan_vbox.pack_start (panning_scroll_box, false, false);
	pan_vbox.pack_start (panning_viewport, false, false);
	pan_vbox.pack_start (panning_link_box, false, false);

	pan_box.pack_start (pan_vbox, true, false);
	pan_vbox.show ();
	pan_box.show ();
	panning_viewport.show ();

	panner = 0;

	meter_metric_area.set_usize (18, -1);
	meter_metric_area.set_name ("MeterMetricsStrip");

	meter_packer.show ();
	gain_slider->show_all ();

	meter_packer.set_spacing (2);
	fader_box.set_spacing (2);

	fader_box.pack_start (*gain_slider, false, false);

	hbox.set_spacing (4);
	hbox.pack_start (fader_box, false, false);
	hbox.pack_start (meter_packer, false, false);

	set_spacing (4);
	pack_start (top_table, false, false);
	pack_start (gain_display_box, false, false);
	pack_start (hbox, false, false);

	show_all ();

	_io.gain_changed.connect (slot (*this, &GainPanMeter::gain_changed));
	_io.panner().Changed.connect (slot (*this, &GainPanMeter::panner_changed));
	_io.panner().LinkStateChanged.connect (slot (*this, &GainPanMeter::update_pan_linkage));

	meter_metric_area.expose_event.connect (slot (*this, &GainPanMeter::meter_metrics_expose));
	gain_adjustment.value_changed.connect (slot (*this, &GainPanMeter::gain_adjusted));
	peak_display.button_release_event.connect (slot (*this, &GainPanMeter::peak_button_release));

	gain_changed (0);
	pan_changed (0);
	update_gain_sensitive ();
	update_pan_sensitive ();
	update_pan_linkage ();
}

gint
GainPanMeter::panning_link_button_press (GdkEventButton* ev)
{
	return stop_signal (panning_link_button, "button-press-event");
}

gint
GainPanMeter::panning_link_button_release (GdkEventButton* ev)
{
	if (!ignore_toggle) {
		_io.panner().set_linked (!_io.panner().linked());
	}
	return TRUE;
}

void
GainPanMeter::panning_link_direction_clicked()
{
	switch (_io.panner().link_direction()) {
	case Panner::SameDirection:
		_io.panner().set_link_direction (Panner::OppositeDirection);
		break;
	default:
		_io.panner().set_link_direction (Panner::SameDirection);
		break;
	}
}

void
GainPanMeter::update_pan_linkage ()
{
	bool x = _io.panner().linked();
	bool bx = panning_link_button.get_active();
	
	if (x != bx) {
		
		ignore_toggle = true;
		panning_link_button.set_active (x);
		ignore_toggle = false;
		
		if (x) {
			static_cast<Gtk::Label*>(panning_link_button.get_child())->set_text (_("link"));
		} else {
			static_cast<Gtk::Label*>(panning_link_button.get_child())->set_text (_("unlink"));
		}
	}

	panning_link_direction_button.set_sensitive (x);

	switch (_io.panner().link_direction()) {
	case Panner::SameDirection:
		static_cast<Pixmap*> (panning_link_direction_button.get_child())->set (forwdblarrow_xpm);
		break;
	default:
		static_cast<Pixmap*> (panning_link_direction_button.get_child())->set (revdblarrow_xpm);
		break;
	}
}

void
GainPanMeter::set_width (Width w)
{
	switch (w) {
	case Wide:
		panning_viewport.set_usize (61, 61);
		if (panner) {
			panner->set_usize (61, 61);
		}
		for (vector<BarController*>::iterator i = pan_bars.begin(); i != pan_bars.end(); ++i) {
				(*i)->set_usize (61, 15);
		}
		static_cast<Gtk::Label*> (panning_link_button.get_child())->set_text (_("link"));
		peak_display_frame.show_all();
		break;
	case Narrow:
		panning_viewport.set_usize (31, 61);
		if (panner) {
			panner->set_usize (31, 61);
		}
		for (vector<BarController*>::iterator i = pan_bars.begin(); i != pan_bars.end(); ++i) {
				(*i)->set_usize (31, 15);
		}
		static_cast<Gtk::Label*> (panning_link_button.get_child())->set_text (_("L"));
		peak_display_frame.hide_all();
		break;
	}

	_width = w;
	setup_meters ();
}

gint
GainPanMeter::meter_metrics_expose (GdkEventExpose *ev)
{
	/* XXX optimize this so that it doesn't do it all everytime */

	double fraction;

	Gdk_Window win (meter_metric_area.get_window());
	Gdk_GC gc (meter_metric_area.get_style()->get_fg_gc (GTK_STATE_NORMAL));
	Gdk_Font font (meter_metric_area.get_style()->get_font());
	gint x, y, width, height, depth;
	gint pos;
	int  db_points[] = { -50, -10, -3, 0, 6 };
	unsigned int i;
	char buf[32];

	win.get_geometry (x, y, width, height, depth);

	for (i = 0; i < sizeof (db_points)/sizeof (db_points[0]); ++i) {
		fraction = log_meter (db_points[i]);
		pos = height - (gint) floor (height * fraction);

		snprintf (buf, sizeof (buf), "%d", db_points[i]);

		gint twidth;
		gint lbearing;
		gint rbearing;
		gint ascent;
		gint descent;

		gdk_string_extents (font,
				    buf,
				    &lbearing,
				    &rbearing,
				    &twidth,
				    &ascent,
				    &descent);

		win.draw_text (font, gc, width - twidth, pos + ascent, buf, strlen (buf));
		// win.draw_line (gc, 0, pos, width, pos);
	}

	return TRUE;
}

GainPanMeter::~GainPanMeter ()
{
	for (vector<Adjustment*>::iterator i = pan_adjustments.begin(); i != pan_adjustments.end(); ++i) {
		delete (*i);
	}
	
	for (vector<BarController*>::iterator i = pan_bars.begin(); i != pan_bars.end(); ++i) {
		delete (*i);
	}

	if (panner) {
		delete panner;
	}

	if (meter_menu) {
		delete meter_menu;
	}

	if (pan_menu) {
		delete pan_menu;
	}

	for (vector<MeterInfo>::iterator i = meters.begin(); i != meters.end(); i++) {
		if ((*i).meter) {
			delete (*i).meter;
		}
	}
}

void
GainPanMeter::update_meters ()
{
	vector<MeterInfo>::iterator i;
	unsigned long n;
	float peak;
	char buf[32];

	for (n = 0, i = meters.begin(); i != meters.end(); ++i, ++n) {
		if ((*i).packed) {
			peak = _io.peak_input_power (n);
			(*i).meter->set (log_meter (_io.peak_input_power (n)));
			if (peak > max_peak) {
				max_peak = peak;
				/* set peak display */
				snprintf (buf, sizeof(buf), "%.1f", max_peak);
				peak_display_label.set_text (buf);
			}
		}
	}
}

void
GainPanMeter::hide_all_meters ()
{
	bool remove_metric_area = false;

	for (vector<MeterInfo>::iterator i = meters.begin(); i != meters.end(); ++i) {
		if ((*i).packed) {
			remove_metric_area = true;
			meter_packer.remove (*((*i).meter));
			(*i).packed = false;
		}
	}

	if (remove_metric_area) {
		if (meter_metric_area.get_parent()) {
			meter_packer.remove (meter_metric_area);
		}
	}
}

void
GainPanMeter::setup_meters ()
{
	unsigned int nmeters;
	guint16 width;

	hide_all_meters ();

	Route* r;

	if ((r = dynamic_cast<Route*> (&_io)) != 0) {
		if (r->meter_pre_fader()) {
			nmeters = r->n_inputs();
		} else {
			nmeters = r->n_outputs();
		}
	} else {
		nmeters = _io.n_outputs();
	}

	if (nmeters == 0) {
		return;
	}

	if (_width == Wide) {
		meter_packer.pack_start (meter_metric_area, false, false);
		meter_metric_area.show_all ();
	}

	if (nmeters <= 2) {
		width = regular_meter_width;
	} else {
		width = thin_meter_width;
	}

	while (meters.size() < nmeters) {
		meters.push_back (MeterInfo());
	}

	for (unsigned int n = 0; n < nmeters; ++n) {
		if (meters[n].width != width) {
			delete meters[n].meter;
			meters[n].meter = new FastMeter ((unsigned long) floor (_session.meter_hold), width, FastMeter::Vertical);
			meters[n].width = width;

			meters[n].meter->add_events (GDK_BUTTON_RELEASE_MASK);
			meters[n].meter->button_release_event.connect
				(bind (slot (*this, &GainPanMeter::meter_button_release), n));
			meters[n].meter->button_release_event.connect_after (slot (do_not_propagate));
		}

		meter_packer.pack_start (*meters[n].meter, false, false);
		meters[n].meter->show_all ();
		meters[n].packed = true;
	}
}	

gint
GainPanMeter::peak_button_release (GdkEventButton* ev)
{
	/* reset peak label */

	max_peak = minus_infinity();
	peak_display_label.set_text (_("-inf"));

	return TRUE;
}

gint
GainPanMeter::meter_button_release (GdkEventButton* ev, unsigned int which)
{
	switch (ev->button) {
	case 1:
		meters[which].meter->clear();
		max_peak = minus_infinity();
		peak_display_label.set_text (_("-inf"));
		break;

	case 3:
		popup_meter_menu (ev);
		break;
	};

	return TRUE;
}

void
GainPanMeter::popup_meter_menu (GdkEventButton *ev)
{
	using namespace Menu_Helpers;

	if (meter_menu == 0) {
		meter_menu = new Gtk::Menu;
		MenuList& items = meter_menu->items();

		items.push_back (MenuElem ("-inf .. +0dBFS"));
		items.push_back (MenuElem ("-10dB .. +0dBFS"));
		items.push_back (MenuElem ("-4 .. +0dBFS"));
		items.push_back (SeparatorElem());
		items.push_back (MenuElem ("-inf .. -2dBFS"));
		items.push_back (MenuElem ("-10dB .. -2dBFS"));
		items.push_back (MenuElem ("-4 .. -2dBFS"));
	}

	meter_menu->popup (1, ev->time);
}

void
GainPanMeter::_gain_printer (char buf[32], Gtk::Adjustment& adj, void *arg)
{
	static_cast<GainPanMeter *>(arg)->gain_printer (buf, adj);
}

void
GainPanMeter::gain_printer (char buf[32], Gtk::Adjustment& adj)
{
	float v = adj.get_value();

	if (v == 0.0) {
		strcpy (buf, _("-inf"));
	} else {
		snprintf (buf, 32, "%.1f", coefficient_to_dB (slider_position_to_gain (v)));
	}
}

void
GainPanMeter::gain_adjusted ()
{
	if (!ignore_toggle) {
		_io.set_gain (slider_position_to_gain (gain_adjustment.get_value()), this);
	}
}

void
GainPanMeter::effective_gain_display ()
{
	gfloat value = gain_to_slider_position (_io.effective_gain());

	if (gain_adjustment.get_value() != value) {
		ignore_toggle = true;
		gain_adjustment.set_value (value);
		ignore_toggle = false;
	}
}

void
GainPanMeter::gain_changed (void *src)
{
	Gtkmmext::UI::instance()->call_slot (slot (*this, &GainPanMeter::effective_gain_display));
}

gint
GainPanMeter::entry_focus_event (GdkEventFocus* ev)
{
	if (ev->in) {
		ARDOUR_UI::instance()->allow_focus (true);
	} else {
		ARDOUR_UI::instance()->allow_focus (false);
	}
	return TRUE;
}

void
GainPanMeter::panner_changed ()
{
	if (Gtkmmext::UI::instance()->caller_is_gui_thread()) {
		setup_pan ();
	} else {
		Gtkmmext::UI::instance()->call_slot (slot (*this, &GainPanMeter::panner_changed));
	}
}

void
GainPanMeter::setup_pan ()
{
	unsigned int nouts = _io.n_outputs ();

	if (nouts == 0) {
		 
	} else if (nouts <= 2) {


		vector<Adjustment*>::size_type asz;
		unsigned int npans = _io.panner().size();

		while (!pan_adjustments.empty()) {
			delete pan_bars.back();
			pan_bars.pop_back ();
			delete pan_adjustments.back();
			pan_adjustments.pop_back ();
		}

		while ((asz = pan_adjustments.size()) < npans) {

			float x;
			BarController* bc;

			/* initialize adjustment with current value of panner */

			_io.panner()[asz]->get_position (x);

			pan_adjustments.push_back (new Adjustment (x, 0, 1.0, 0.05, 0.1));
			pan_adjustments.back()->value_changed.connect (bind (slot (*this, &GainPanMeter::pan_adjustment_changed), (unsigned long) asz));

			_io.panner()[asz]->Changed.connect (bind (slot (*this, &GainPanMeter::pan_value_changed), (unsigned long) asz));

			bc = new BarController (*pan_adjustments[asz], 
						&_io.panner()[asz]->midi_control(),
						bind (slot (*this, &GainPanMeter::pan_printer), pan_adjustments[asz]));

			bc->set_name ("PanSlider");
			bc->set_shadow_type (GTK_SHADOW_NONE);
			bc->set_style (BarController::Line);
			bc->get_spin_button().focus_in_event.connect (slot (*this, &GainPanMeter::entry_focus_event));
			bc->get_spin_button().focus_out_event.connect (slot (*this, &GainPanMeter::entry_focus_event));

			bc->StartGesture.connect (bind (slot (_io, &IO::start_pan_touch), (unsigned long) asz));
			bc->StopGesture.connect (bind (slot (_io, &IO::end_pan_touch), (unsigned long) asz));

			char buf[64];
			snprintf (buf, sizeof (buf), _("panner for channel %u"), asz + 1);
			ARDOUR_UI::instance()->tooltips().set_tip (bc->event_widget(), buf);

			bc->event_widget().button_release_event.connect
				(bind (slot (*this, &GainPanMeter::pan_button_event), (unsigned long) asz));

			pan_bars.push_back (bc);
			switch (_width) {
			case Wide:
				pan_bars.back()->set_usize (61, 15);
				break;
			case Narrow:
				pan_bars.back()->set_usize (31, 15);
				break;
			}

			pan_bar_packer.pack_start (*pan_bars.back(), false, false);
		}

		/* now that we actually have the pan bars,
		   set their sensitivity based on current
		   automation state.
		*/

		update_pan_sensitive ();

		panning_viewport.remove ();
		panning_viewport.add (pan_bar_packer);
		panning_viewport.show_all ();

	} else {

		int w;

		switch (_width) {
		case Wide:
			w = 61;
			break;
		case Narrow:
			w = 31;
			break;
		}

		if (panner == 0) {
			panner = new Panner2d (w, 61);
			panner->set_name ("MixerPanZone");
			panner->show ();

			panner->PuckMoved.connect (slot (*this, &GainPanMeter::panner_moved));
			panner->BypassToggled.connect (slot (*this, &GainPanMeter::panner_bypass_toggled));
		}

		update_pan_sensitive ();
		panner->set_usize (w, 61);

		/* add pucks */

		panner->drop_pucks ();

		switch (_io.n_inputs()) {
		case 0:
			break;

		case 1:
			panner->add_puck ("", 0.0f, 0.5f);
			break;

		case 2:
			panner->add_puck ("L", 0.5f, 0.25f);
			panner->add_puck ("R", 0.25f, 0.5f);
			panner->show_puck (0);
			panner->show_puck (1);
			break;

		default:
			for (unsigned long i = 0; i < _io.n_inputs(); ++i) {
				char buf[64];
				snprintf (buf, sizeof (buf), "%lu", i);
				panner->add_puck (buf, 0.0f, 0.5f);
				panner->show_puck (i);
			}
			break;
		}

		/* add all outputs */

		panner->drop_targets ();

		for (unsigned int n = 0; n < _io.panner().nouts(); ++n) {
			panner->add_target (_io.panner().output (n).x,
					    _io.panner().output (n).y);
		}

		panner->allow_x_motion (true);
		panner->allow_y_motion (true);
		panner->allow_target_motion (true);

#if 0
		/* if all pans are 1.0 then pans are considered bypassed */
		bool byp = true;
		for (unsigned long n = 0; n < nouts; ++n) {
			if (_io.pans()[n] != 1.0) {
				byp = false;
				break;
			}
		}
		panner->set_bypassed (byp);
#endif	

		/* and finally, add it to the panner frame */

		panning_viewport.remove ();
		panning_viewport.add (*panner);
		panning_viewport.show_all ();
	}
}

gint
GainPanMeter::pan_button_event (GdkEventButton* ev, unsigned long which)
{
	switch (ev->button) {
	case 3:
		if (pan_menu == 0) {
			pan_menu = manage (new Menu);
		}
		build_pan_menu (which);
		pan_menu->popup (1, ev->time);
		return TRUE;
		break;
	default:
		return FALSE;
	}

	return FALSE; // what's wrong with gcc?
}

void
GainPanMeter::build_pan_menu (unsigned long which)
{
	using namespace Menu_Helpers;
	MenuList& items (pan_menu->items());

	items.clear ();

	items.push_back (CheckMenuElem ("Mute", bind (slot (*this, &GainPanMeter::pan_mute), which)));
	(dynamic_cast<CheckMenuItem*> (items.back()))->set_active (_io.panner()[which]->muted());

	items.push_back (MenuElem ("Reset", slot (*this, &GainPanMeter::pan_reset)));
	items.push_back (SeparatorElem());
	items.push_back (MenuElem ("Reset all"));
}

void
GainPanMeter::pan_mute (unsigned long which)
{
	StreamPanner* sp = _io.panner()[which];
	sp->set_muted (!sp->muted());
}

void
GainPanMeter::pan_reset ()
{
}

void
GainPanMeter::effective_pan_display ()
{
	if (_io.panner().empty()) {
		return;
	}

	switch (_io.n_outputs()) {
	case 0: 
	case 1:
		/* relax */
		break;

	case 2:
		update_pan_bars (true);
		break;

	default:
                //panner->move_puck (pan_value (v, right), 0.5);
		break;
	}
}

void
GainPanMeter::pan_changed (void *src)
{
	if (src == this) {
		return;
	}

	if (_io.panner().size() == 0) {
		panning_link_button.hide ();
		pan_box.hide ();
		return;
	} else {
		panning_link_button.show ();
		pan_box.show ();
	}

	unsigned long nouts = _io.n_outputs();

	switch (nouts) {
	case 0:
	case 1:
		/* relax */
		break;

	case 2:
		update_pan_bars (false);
		break;

	default:
		// panner->move_puck (pan_value (pans[0], pans[1]), 0.5);
		break;
	}
}

void
GainPanMeter::pan_adjustment_changed (unsigned long which)
{
	if (!in_pan_update && which < _io.panner().size()) {

		float xpos;
		float val = pan_adjustments[which]->get_value ();
		_io.panner()[which]->get_position (xpos);

		/* add a kinda-sorta detent for the middle */
		
		if (val != 0.5 && Panner::equivalent (val, 0.5)) {
			/* this is going to be reentrant, so just 
			   return after it.
			*/

			in_pan_update = true;
			pan_adjustments[which]->set_value (0.5);
			in_pan_update = false;
			return;
		}
		
		if (!Panner::equivalent (val, xpos)) {

			_io.panner()[which]->set_position (val);
			/* XXX 
			   the panner objects have no access to the session,
			   so do this here. ick.
			*/
			_session.set_dirty();
		}
	}
}

void
GainPanMeter::pan_value_changed (unsigned long which)
{
	if (!Gtkmmext::UI::instance()->caller_is_gui_thread()) {
		Gtkmmext::UI::instance()->call_slot (bind (slot (*this, &GainPanMeter::pan_value_changed), which));
		return;
	}
							   
	if (which < _io.panner().size()) {
		float xpos;
		float val = pan_adjustments[which]->get_value ();

		_io.panner()[which]->get_position (xpos);

		if (!Panner::equivalent (val, xpos)) {
			in_pan_update = true;
			pan_adjustments[which]->set_value (xpos);
			in_pan_update = false;
		}
	}
}	

void
GainPanMeter::update_pan_bars (bool only_if_aplay)
{
	unsigned long n;
	vector<Adjustment*>::iterator i;

	in_pan_update = true;

	/* this runs during automation playback, and moves the bar controllers
	   and/or pucks around.
	*/

	for (i = pan_adjustments.begin(), n = 0; i != pan_adjustments.end(); ++i, ++n) {
		float xpos, val;

		if (only_if_aplay) {
			AutomationList& alist (_io.panner()[n]->automation());
			
			if (!alist.automation_playback()) {
				continue;
			}
		}

		_io.panner()[n]->get_effective_position (xpos);
		val = (*i)->get_value ();
		
		if (!Panner::equivalent (val, xpos)) {
			(*i)->set_value (xpos);
		}
	}

	in_pan_update = false;
}

void
GainPanMeter::pan_printer (char *buf, unsigned int len, Adjustment* adj)
{
	float val = adj->get_value();

	if (val == 0.0f) {
		snprintf (buf, len, X_("L"));
	} else if (val == 1.0f) {
		snprintf (buf, len, X_("R"));
	} else if (Panner::equivalent (val, 0.5f)) {
		snprintf (buf, len, X_("C"));
	} else {
		/* don't print anything */
		buf[0] = '\0';
	}
}


void
GainPanMeter::panner_moved (int which)
{
	if (panner->bypassed()) {
		return;
	}

	float x, y;
	panner->puck_position (which, x, y);

	_io.panner()[which]->set_position (x, y);
}

void
GainPanMeter::panner_bypass_toggled ()
{
#if 0
	vector<pan_t> pans;
	unsigned long nouts = _io.n_outputs();
	unsigned long nins = _io.n_inputs();

	if (panner->bypassed()) {
		/* all pans to full gain */
		for (unsigned int i=0; i < nouts; ++i) {
			pans.push_back (1.0);
		}
		_io.set_pans (pans, this);
	}
	else if (nins <= 1) {
		// reset if not more than one input
		float scale = 1.0 / nouts;
		for (unsigned int i=0; i < nouts; ++i) {
			pans.push_back (scale);
		}
		_io.set_pans (pans, this);
		
		panner_moved();
	}
#endif
}

void
GainPanMeter::set_meter_strip_name (string name)
{
	meter_metric_area.set_name (name);
}

void
GainPanMeter::update_gain_sensitive ()
{
	static_cast<Gtkmmext::SliderController*>(gain_slider)->set_sensitive (!(_io.gain_automation_state() & Play));
}

void
GainPanMeter::update_pan_sensitive ()
{
	bool sensitive = !(_io.panner().automation_state() & Play);

	switch (_io.n_outputs()) {
	case 0:
	case 1:
		break;
	case 2:
		for (vector<BarController*>::iterator i = pan_bars.begin(); i != pan_bars.end(); ++i) {
			(*i)->set_sensitive (sensitive);
		}
		break;
	default:
		if (panner) {
			panner->set_sensitive (sensitive);
		}
		break;
	}
}

void
GainPanMeter::meter_pre_toggled ()
{
	Route* r;

	if ((r = dynamic_cast<Route*> (&_io)) != 0) {
		bool x = meter_pre_button.get_active ();

		if (x != r->meter_pre_fader()) {
			r->set_meter_pre_fader (x, this);
		}
	}
}

void
GainPanMeter::meter_changed (void *src)
{
	Route* r;

	if ((r = dynamic_cast<Route*> (&_io)) != 0) {
		bool x = r->meter_pre_fader();
		if (x != meter_pre_button.get_active()) {
			meter_pre_button.set_active (x);
		}
		setup_meters ();
	}
}

gint
GainPanMeter::start_gain_touch (GdkEventButton* ev)
{
	_io.start_gain_touch ();
	return FALSE;
}

gint
GainPanMeter::end_gain_touch (GdkEventButton* ev)
{
	_io.end_gain_touch ();
	return FALSE;
}
