/*
    Copyright (C) 2002 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: audio_track.cc,v 1.42 2004/06/25 15:34:34 pauld Exp $
*/

#include <ardour/audio_track.h>
#include <ardour/diskstream.h>
#include <ardour/session.h>
#include <ardour/redirect.h>
#include <ardour/route_group_specialized.h>
#include <ardour/insert.h>
#include <ardour/audioplaylist.h>
#include <ardour/panner.h>

#include "i18n.h"

using namespace std;
using namespace SigC;
using namespace ARDOUR;

AudioTrack::AudioTrack (Session& sess, string name, Route::Flag flag)
	: Route (sess, name, 1, -1, -1, -1, flag),
	  diskstream (0)
{
	DiskStream::Flag dflags = DiskStream::Flag (0);

	if (_flags & Hidden) {
		dflags = DiskStream::Flag (dflags | DiskStream::Hidden);
	} else {
		dflags = DiskStream::Flag (dflags | DiskStream::Recordable);
	}

	DiskStream* ds = new DiskStream (_session, name, dflags);
	
	set_diskstream (*ds, this);
	
	_freeze_state = NoFreeze;

	_declickable = true;
	
	output_changed.connect (slot (*this, &AudioTrack::output_change_handler));
}

AudioTrack::AudioTrack (Session& sess, const XMLNode& node)
	: Route (sess, "to be renamed", 0, 0, -1, -1),
	  diskstream (0)
{
	set_state (node);
	_freeze_state = NoFreeze;
	_declickable = true;
	output_changed.connect (slot (*this, &AudioTrack::output_change_handler));
}

AudioTrack::~AudioTrack ()
{
	if (diskstream) {
		diskstream->unref();
	}
}

int
AudioTrack::deprecated_use_diskstream_connections ()
{
	if (diskstream->deprecated_io_node == 0) {
		return 0;
	}

	const XMLProperty* prop;
	XMLNode& node (*diskstream->deprecated_io_node);

	/* don't do this more than once. */

	diskstream->deprecated_io_node = 0;

	set_input_minimum (-1);
	set_input_maximum (-1);
	set_output_minimum (-1);
	set_output_minimum (-1);
	
	if ((prop = node.property ("gain")) != 0) {
		set_gain (atof (prop->value().c_str()), this);
		_gain = _desired_gain;
	}

	if ((prop = node.property ("input-connection")) != 0) {
		Connection* c = _session.connection_by_name (prop->value());
		
		if (c == 0) {
			error << compose(_("Unknown connection \"%1\" listed for input of %2"), prop, _name) << endmsg;
			return -1;
		} else {
			use_input_connection (*c, this);
		}

	} else if ((prop = node.property ("inputs")) != 0) {
		if (set_inputs (prop->value())) {
			error << compose(_("improper input channel list in XML node (%1)"), prop->value()) << endmsg;
			return -1;
		}
	}
	
	return 0;
}

int
AudioTrack::set_diskstream (DiskStream& ds, void *src)
{
	if (diskstream) {
		diskstream->unref();
	}

	diskstream = &ds.ref();
	diskstream->set_io (*this);

	if (diskstream->deprecated_io_node) {

		if (!connecting_legal) {
			ConnectingLegal.connect (slot (*this, &AudioTrack::deprecated_use_diskstream_connections));
		} else {
			deprecated_use_diskstream_connections ();
		}
	}

	diskstream->set_record_enabled (false, this);
	diskstream->monitor_input (false);

	ic_connection.disconnect();
	icc_connection.disconnect();

	ic_connection = input_changed.connect (slot (*diskstream, &DiskStream::handle_input_change));
	icc_connection = input_configuration_changed.connect (slot (*diskstream, &DiskStream::handle_input_configuration_change));

	diskstream_changed (src); /* EMIT SIGNAL */

	return 0;
}	

int 
AudioTrack::use_diskstream (string name)
{
	DiskStream *dstream;

	if ((dstream = _session.diskstream_by_name (name)) == 0) {
		error << compose(_("AudioTrack: diskstream \"%1\" not known by session"), name) << endmsg;
		return -1;
	}
	
	return set_diskstream (*dstream, this);
}

int 
AudioTrack::use_diskstream (id_t id)
{
	DiskStream *dstream;

	if ((dstream = _session.diskstream_by_id (id)) == 0) {
		error << compose(_("AudioTrack: diskstream \"%1\" not known by session"), id) << endmsg;
		return -1;
	}
	
	return set_diskstream (*dstream, this);
}

bool
AudioTrack::record_enabled () const
{
	return diskstream->record_enabled();
}

void
AudioTrack::set_record_enable (bool yn, void *src)
{
	if (_mix_group && src != _mix_group && _mix_group->is_active()) {
		_mix_group->apply (&AudioTrack::set_record_enable, yn, _mix_group);
		return;
	}

	diskstream->set_record_enabled (yn, src);
}

XMLNode&
AudioTrack::state(bool full_state)
{
	XMLNode& track_state (Route::state (full_state));
	char buf[64];

	/* we don't return diskstream state because we don't
	   own the diskstream exclusively. control of the diskstream
	   state is ceded to the Session, even if we create the
	   diskstream.
	*/

	snprintf (buf, sizeof (buf), "%Lu", diskstream->id());
	track_state.add_property ("diskstream-id", buf);

	return track_state;
}

int
AudioTrack::set_state (const XMLNode& node)
{
	const XMLProperty *prop;

	if (Route::set_state (node)) {
		return -1;
	}

	if ((prop = node.property ("diskstream-id")) == 0) {
		
		/* some old sessions use the diskstream name rather than the ID */

		if ((prop = node.property ("diskstream")) == 0) {
			fatal << _("programming error: AudioTrack given state without diskstream!") << endmsg;
			/*NOTREACHED*/
			return -1;
		}

		return use_diskstream (prop->value());
	}

 	id_t id = strtoull (prop->value().c_str(), 0, 10);
	return use_diskstream (id);
}

void
AudioTrack::passthru_silence (jack_nframes_t start_frame, jack_nframes_t end_frame, jack_nframes_t nframes, jack_nframes_t offset, int declick)
{
	vector<Sample*> bufs;

	get_process_buffers (bufs);

	/* XXX can we optimize this to use a set of silent buffers? */

	for (vector<Sample*>::iterator o = bufs.begin(); o != bufs.end(); ++o) {
		memset ((*o), 0, sizeof (Sample) * nframes);
	}
	
	process_output_buffers (bufs, start_frame, end_frame, nframes, offset, true, declick, false);
}

int 
AudioTrack::no_roll (jack_nframes_t nframes, jack_nframes_t start_frame, jack_nframes_t end_frame, jack_nframes_t offset, 
		     bool session_state_changing, bool can_record, bool rec_monitors_input)
{
	if (n_outputs() == 0) {
		return 0;
	}

	if (!_active) {
		silence (nframes, offset);
	}

	if (session_state_changing) {

		/* XXX is this safe to do against transport state changes? */

		passthru_silence (start_frame, end_frame, nframes, offset, 0);
		return 0;
	}
	
	if (_session.get_auto_input()) {
		if (Config->get_no_sw_monitoring()) {
			passthru_silence (start_frame, end_frame, nframes, offset, 0);
		} else {
			passthru (start_frame, end_frame, nframes, offset, 0);
		}
	} else {
		if (diskstream->record_enabled()) {
			if (Config->get_no_sw_monitoring()) {
				just_meter_input (start_frame, end_frame, nframes, offset);
				passthru_silence (start_frame, end_frame, nframes, offset, 0);
			} else {
				passthru (start_frame, end_frame, nframes, offset, 0);
			}
		} else {
			passthru_silence (start_frame, end_frame, nframes, offset, 0);
		}
	}

	return 0;
}

int
AudioTrack::roll (jack_nframes_t nframes, jack_nframes_t start_frame, jack_nframes_t end_frame, jack_nframes_t offset, int declick,
		  bool can_record, bool rec_monitors_input)
{
	int dret;
	Sample* b;
	Sample* tmpb;

	automation_snapshot (start_frame);

	if (n_outputs() == 0 && _redirects.empty()) {
		return 0;
	}

	if (!_active) {
		silence (nframes, offset);
		return 0;
	}

	if ((nframes = check_initial_delay (nframes, offset)) == 0) {
		/* need to do this so that the diskstream sets its
		   playback distance to zero, thus causing diskstream::commit
		   to do nothing.
		*/
		return diskstream->process (0, 0, false, false);
	}

	_silent = false;
	apply_gain_automation = false;
	
	if ((dret = diskstream->process (nframes, offset, can_record, rec_monitors_input)) != 0) {

		silence (nframes, offset);

		/* positive error code means that the diskstream lock could not be taken */

		if (dret < 0) {
			error << compose(_("%1: disk stream error at frame %2 (%3)"), diskstream->name(), start_frame, dret) 
			      << endmsg;
		}

		return dret;
	}

	if (diskstream->record_enabled() && !can_record && !_session.get_auto_input()) {

		/* not actually recording, but we want to hear the input material anyway */
		passthru (start_frame, end_frame, nframes, offset, 0);

	} else if ((b = diskstream->playback_buffer(0)) != 0) {

		/*
		  XXX is it true that the earlier test on n_outputs()
		  means that we can avoid checking it again here? i think
		  so, because changing the i/o configuration of an IO
		  requires holding the AudioEngine lock, which we hold
		  while in the process() tree.
		*/

		/* copy the diskstream data to all output buffers */
		
		vector<Sample*> bufs;

		get_process_buffers (bufs);
		
		unsigned int n=1;
		for (vector<Sample*>::iterator i = bufs.begin(); i != bufs.end(); ++i, ++n) {
			memcpy ((*i), b, sizeof (Sample) * nframes); 
			if (n < diskstream->n_channels()) {
				tmpb = diskstream->playback_buffer(n);
				if (tmpb!=0) {
					b = tmpb;
				}
			}
		}

		/* don't waste time with automation if we're recording */

		if (!diskstream->record_enabled()) {
			
			if (gain_automation_playback()) {
				apply_gain_automation = _gain_automation_curve.rt_safe_get_vector (start_frame, end_frame, _session.gain_automation_buffer(), nframes);
			}
		}

		process_output_buffers (bufs, start_frame, end_frame, nframes, offset, (!_session.get_record_enabled() || _session.get_recording_plugins()), declick, true);
		
	} else {
		/* problem with the diskstream; just be quiet for a bit */
		silence (nframes, offset);
	}

	return 0;
}

void
AudioTrack::toggle_monitor_input ()
{
	for (vector<Port*>::iterator i = _inputs.begin(); i != _inputs.end(); ++i) {
		cerr << "toggle monitor on "
		     << (*i)->name()
		     << " to " 
		     << !(*i)->monitoring_input()
		     << endl;

		(*i)->request_monitor_input(!(*i)->monitoring_input());
	}
}

void
AudioTrack::output_change_handler (void *src)
{
}

void
AudioTrack::set_name (string str, void *src)
{
	diskstream->set_name (str, src);
	IO::set_name (str, src);
}

int
AudioTrack::export_stuff (vector<Sample*>& buffers, jack_nframes_t start, jack_nframes_t nframes)
{
	gain_t  gain_automation[nframes];
	gain_t  gain_buffer[nframes];
	float   mix_buffer[nframes];
	RedirectList::iterator i;
	bool post_fader_work = false;
	gain_t this_gain = _gain;
	vector<Sample*>::iterator bi;
	Sample * b;
	
	LockMonitor rlock (redirect_lock, __LINE__, __FILE__);
		
	if (diskstream->playlist()->read (buffers[0], mix_buffer, gain_buffer, start, nframes) != nframes) {
		return -1;
	}

	unsigned int n=1;
	bi = buffers.begin();
	b = buffers[0];
	++bi;
	for (; bi != buffers.end(); ++bi, ++n) {
		if (n < diskstream->n_channels()) {
			if (diskstream->playlist()->read ((*bi), mix_buffer, gain_buffer, start, nframes, n) != nframes) {
				return -1;
			}
			b = (*bi);
		}
		else {
			/* duplicate last across remaining buffers */
			memcpy ((*bi), b, sizeof (Sample) * nframes); 
		}
	}


	/* note: only run inserts during export. other layers in the machinery
	   will already have checked that there are no external port inserts.
	*/
	
	for (i = _redirects.begin(); i != _redirects.end(); ++i) {
		Insert *insert;
		
		if ((insert = dynamic_cast<Insert*>(*i)) != 0) {
			switch (insert->placement()) {
			case PreFader:
				insert->run (buffers, nframes, 0);
				break;
			case PostFader:
				post_fader_work = true;
				break;
			}
		}
	}
	
	if (_gain_automation_curve.automation_state() == Play) {
		
		_gain_automation_curve.get_vector (start, start + nframes, gain_automation, nframes);

		for (bi = buffers.begin(); bi != buffers.end(); ++bi) {
			Sample *b = *bi;
			for (jack_nframes_t n = 0; n < nframes; ++n) {
				b[n] *= gain_automation[n];
			}
		}

	} else {

		for (bi = buffers.begin(); bi != buffers.end(); ++bi) {
			Sample *b = *bi;
			for (jack_nframes_t n = 0; n < nframes; ++n) {
				b[n] *= this_gain;
			}
		}
	}

	if (post_fader_work) {

		for (i = _redirects.begin(); i != _redirects.end(); ++i) {
			PluginInsert *insert;
			
			if ((insert = dynamic_cast<PluginInsert*>(*i)) != 0) {
				switch ((*i)->placement()) {
				case PreFader:
					break;
				case PostFader:
					insert->run (buffers, nframes, 0);
					break;
				}
			}
		}
	} 

	return 0;
}

jack_nframes_t
AudioTrack::update_total_latency ()
{
	_own_latency = 0;

	for (RedirectList::iterator i = _redirects.begin(); i != _redirects.end(); ++i) {
		if ((*i)->active ()) {
			_own_latency += (*i)->latency ();
		}
	}

	set_port_latency (_own_latency);

	return _own_latency;
}

void
AudioTrack::freeze (UndoCommand fc)
{
	_freeze_cmd = fc;
	_freeze_state = Frozen;
}

void
AudioTrack::unfreeze ()
{
	if (_freeze_state == Frozen) {
		_freeze_cmd.undo ();
		_freeze_state = UnFrozen;
	}
}

void
AudioTrack::refreeze ()
{
	if (_freeze_state == UnFrozen) {
		_freeze_cmd.redo ();
		_freeze_state = Frozen;
	}
}

void
AudioTrack::get_process_buffers (vector<Sample*>& bufs)
{
	unsigned long limit = max ((unsigned long) diskstream->n_channels(), redirect_max_outs);
	
	if (n_outputs() == 1) {
		cerr << _name << " needs " 
		     << limit 
		     << " buffers c/o "
		     << diskstream->n_channels()
		     << " and "
		     << redirect_max_outs
		     << endl;
	}

	/* we have to do this to make sure we only pass the correct number
	   of buffers to the processing chain. 

	   XXX maybe fix me.
	*/

	for (unsigned long n = 0; n < limit; ++n) {
		bufs.push_back (_session.get_passthru_buffers()[n]);
	}
}
