/*
    Copyright (C) 2002-4 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: mtc_slave.cc,v 1.9 2004/06/11 16:08:14 pauld Exp $
*/

#include <errno.h>
#include <poll.h>
#include <sys/types.h>
#include <unistd.h>
#include <pbd/error.h>
#include <pbd/failed_constructor.h>
#include <pbd/pthread_utils.h>

#include <midi++/port.h>
#include <ardour/slave.h>
#include <ardour/session.h>
#include <ardour/audioengine.h>
#include <ardour/cycles.h>

#include "i18n.h"

using namespace ARDOUR;
using namespace SigC;
using namespace MIDI;

MTC_Slave::MTC_Slave (Session& s, MIDI::Port& p) 
	: session (s),
	  port (p)
{
	last_inbound_mtc_frame = 0;
	mtc_speed_counter = 0;
	mtc_speed = 0;
	_mtc_skip_limit = 0; /* stop at any missed MTC quarter frame */
	_mtc_skip_count = 0;
	have_first_accumulated_mtc_speed = false;

	memset (mtc_speed_accumulator, 0, sizeof (mtc_speed_accumulator));

	port.input()->mtc_time.connect (slot (*this, &MTC_Slave::update_mtc_time));
	port.input()->mtc_status.connect (slot (*this, &MTC_Slave::update_mtc_status));
	port.input()->mtc_skipped.connect (slot (*this, &MTC_Slave::mtc_skip));
}

MTC_Slave::~MTC_Slave()
{
}

bool
MTC_Slave::mtc_skip ()
{
	if (_mtc_skip_count++ >= _mtc_skip_limit) {
		cerr << "--- one skip beyond! count = " << _mtc_skip_count << " limit = " << _mtc_skip_limit  << endl;
		return false; /* this skip is too much. this will cause MTC to "stop" */
	} 
	/* this skip is OK, just keep rolling along */
	cerr << "--- skipped " << _mtc_skip_count << " of " << _mtc_skip_limit << " keep on movin'" << endl;
	return true;
}

void
MTC_Slave::set_skip_limit (unsigned long n)
{
	_mtc_skip_limit = n;
	_mtc_skip_count = 0;
}

void
MTC_Slave::update_mtc_time (const byte *msg, bool was_full)
{
	jack_nframes_t now = session.engine().frame_time();
	jack_nframes_t frame_rate = session.frame_rate();
	jack_nframes_t mtc_frame;

	mtc_frame = (msg[3] * 60 * 60 * frame_rate) + 
		(msg[2] * 60 * frame_rate) +
		(msg[1] * frame_rate) +
		(msg[0] * session.frames_per_smpte_frame());

	// fprintf (stderr, "MTC: %02d:%02d:%02d:%02d\n", msg[3], msg[2], msg[1], msg[0]);
	SMPTE_Time st;
	session.smpte_time (session.transport_frame(), st);
	// fprintf (stderr, "AD@%lu %02d:%02d:%02d:%02d\n", session.transport_frame(), st.hours, st.minutes, st.seconds, st.frames);
	
	if (was_full) {

		mtc_speed_counter = 0;
		memset (mtc_speed_accumulator, 0, sizeof (mtc_speed_accumulator));

		current.guard1++;
		current.timestamp = 0;
		current.guard2++;

		SMPTE_Time st;
		session.smpte_time (mtc_frame, st);
		cerr << "LOCATE to: " << mtc_frame << " = ";
		fprintf (stderr, "AD@%lu %02d:%02d:%02d:%02d\n", session.transport_frame(), st.hours, st.minutes, st.seconds, st.frames);
		
		session.request_locate (mtc_frame, false);

	} else {

		if (current.timestamp) {

			jack_nframes_t mtc_delta;
			jack_nframes_t wall_clock_frames;
			float mtc_apparent_speed;

			if (mtc_frame > last_inbound_mtc_frame) {
				mtc_delta = mtc_frame - last_inbound_mtc_frame;
			} else {
				mtc_delta = last_inbound_mtc_frame - mtc_frame;
			}
			
			wall_clock_frames = now - current.timestamp;

			if (!have_first_accumulated_mtc_speed) {
				mtc_speed = (0.8 * mtc_speed) + (0.2 * ((float) mtc_delta / wall_clock_frames));
			}

			mtc_apparent_speed = (float) mtc_delta / wall_clock_frames;
			
			if (mtc_apparent_speed < 0.98 || mtc_apparent_speed > 1.02) {
				cerr << "*** CORRECTED ABERRANT SPEED " << mtc_apparent_speed << endl;
				mtc_apparent_speed = mtc_speed;
			}
	
			/* stuff the current speed into the MTC speed accumulator.
			   we'll average later.
			*/
			    
			if (fabs (mtc_apparent_speed) > 0.0f) {
				
				mtc_speed_accumulator[mtc_speed_counter++] = mtc_apparent_speed;
			}

			if (mtc_speed_counter >= mtc_accumulator_size) {
				
				float total = 0;
				
				for (int i = 0; i < mtc_accumulator_size; ++i) {
					total += mtc_speed_accumulator[i];
				}
				
				mtc_speed = total / mtc_accumulator_size;
				
				if (mtc_apparent_speed > 0.98 && mtc_apparent_speed < 1.02) {
					mtc_speed = 1.0;
				}

				mtc_speed_counter = 0;
				memset (mtc_speed_accumulator, 0, sizeof (mtc_speed_accumulator));
					have_first_accumulated_mtc_speed = true;
			}
		}
		
		current.guard1++;
		// XXX need memory barrier here
		current.mtc.frames = msg[0];
		current.mtc.seconds = msg[1];
		current.mtc.minutes = msg[2];
		current.mtc.hours = msg[3];
		current.timestamp = now;
		// XXX need memory barrier here
		current.guard2++;
		
		last_inbound_mtc_frame = mtc_frame;
	}
}

void
MTC_Slave::update_mtc_status (MIDI::Parser::MTC_Status status)
{
	cerr << "received new MTC status\n";

	switch (status) {
	case MTC_Stopped:
		cerr << "MTC stopped ...\n";
		_mtc_skip_count = 0;
		session.request_transport_speed (0);
		break;

	case MTC_Forward:
		cerr << "MTC starts forward ...\n";
		_mtc_skip_count = 0;
		session.request_transport_speed (1.0);
		break;

	case MTC_Backward:
		cerr << "MTC starts backward ...\n";
		_mtc_skip_count = 0;
		session.request_transport_speed (-1.0);
		break;
	}
}

void
MTC_Slave::read_current (SafeTime *st) const
{
	int tries = 0;
	do {
		if (tries == 10) {
			error << _("MTC Slave: atomic read of current time failed, sleeping!") << endmsg;
			usleep (20);
			tries = 0;
		}

		*st = current;
		tries++;

	} while (st->guard1 != st->guard2);
}

jack_nframes_t
MTC_Slave::current_audio_frame ()
{
	SafeTime last;
	jack_nframes_t frame_rate = session.frame_rate();
	unsigned long elapsed;

	read_current (&last);

	jack_nframes_t now = session.engine().frame_time();

	/* scale elapsed time by the current MTC speed */

	if (now - last.timestamp > frame_rate/2) {
		/* no updates: signal stopped */
		_stopped = true;
		elapsed = 0;
	} else {
		_stopped = false;
		elapsed = (unsigned long) floor (mtc_speed * (now - last.timestamp));
	}

	/* now add the most recent timecode value plus the estimated elapsed interval */

	return elapsed +(last.mtc.hours * 60 * 60 * frame_rate) + 
		(last.mtc.minutes * 60 * frame_rate) +
		(last.mtc.seconds * frame_rate) +
		(last.mtc.frames * session.frames_per_smpte_frame());
}

bool
MTC_Slave::locked () const
{
	return port.input()->mtc_locked();
}

bool
MTC_Slave::ok() const
{
	return true;
}

float
MTC_Slave::current_speed ()
{
	return mtc_speed;
}
