/*
    Copyright (C) 2000-2002 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: plugin.cc,v 1.2 2004/04/12 04:26:20 pauld Exp $
*/

#include <vector>
#include <string>

#include <cstdlib>
#include <cstdio> // so libraptor doesn't complain
#include <cmath>
#include <dirent.h>
#include <sys/stat.h>
#include <cerrno>

#include <lrdf.h>

#include <pbd/compose.h>
#include <pbd/error.h>
#include <pbd/pathscanner.h>
#include <pbd/xml++.h>

#include <midi++/manager.h>

#include <ardour/ardour.h>
#include <ardour/session.h>
#include <ardour/audioengine.h>
#include <ardour/plugin.h>

#include <pbd/stl_delete.h>

#include "i18n.h"
#include <locale.h>

using namespace ARDOUR;

Plugin::Plugin (AudioEngine& e, Session& s)
	: _engine (e), _session (s)
{
}

Plugin::Plugin (const Plugin& other)
	: _engine (other._engine), _session (other._session), _info (other._info)
{
}

void
Plugin::setup_midi_controls ()
{
	unsigned long port_cnt;

	port_cnt = parameter_count();

	/* set up a vector of null pointers for the MIDI controls.
	   we'll fill this in on an as-needed basis.
	*/

	for (unsigned long i = 0; i < port_cnt; ++i) {
		midi_controls.push_back (0);
	}
}

Plugin::~Plugin ()
{
	for (vector<MIDIPortControl*>::iterator i = midi_controls.begin(); i != midi_controls.end(); ++i) {
		if (*i) {
			delete *i;
		}
	}
}

MIDI::Controllable *
Plugin::get_nth_midi_control (unsigned long n)
{
	if (n >= parameter_count()) {
		return 0;
	}

	if (midi_controls[n] == 0) {

		Plugin::ParameterDescriptor desc;

		get_parameter_descriptor (n, desc);

		midi_controls[n] = new MIDIPortControl (*this, n, 0, desc.lower, desc.upper, desc.toggled, desc.logarithmic);
	}
	
	return midi_controls[n];
}

Plugin::MIDIPortControl::MIDIPortControl (Plugin& p, unsigned long port_id, MIDI::Port *port,
					  float low, float up, bool t, bool loga)
	: MIDI::Controllable (port, 0), plugin (p), absolute_port (port_id)
{
	toggled = t;
	logarithmic = loga;
	lower = low;
	upper = up;
	range = upper - lower;
}

void
Plugin::MIDIPortControl::set_value (float value)
{
	if (toggled) {
		if (value > 0.5) {
			value = 1.0;
		} else {
			value = 0.0;
		}
	} else {
		value = lower + (range * value);
		
		if (logarithmic) {
			value = exp(value);
		}
	}

	plugin.set_parameter (absolute_port, value);
}

void
Plugin::reset_midi_control (MIDI::Port* port, bool on)
{
	MIDI::channel_t chn;
	MIDI::eventType ev;
	MIDI::byte extra;

	if (!on) {
		port = 0;
	}

	for (vector<MIDIPortControl*>::iterator i = midi_controls.begin(); i != midi_controls.end(); ++i) {
		if (*i == 0)
			continue;
		(*i)->get_control_info (chn, ev, extra);
		(*i)->midi_rebind (port, chn);
	}
}

