/*
    Copyright (C) 1999-2002 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: session_process.cc,v 1.44 2004/06/29 17:48:09 pauld Exp $
*/

#include <cmath>
#include <cerrno>
#include <unistd.h>

#include <ardour/timestamps.h>

#include <pbd/error.h>
#include <pbd/lockmonitor.h>

#include <ardour/ardour.h>
#include <ardour/session.h>
#include <ardour/diskstream.h>
#include <ardour/audioengine.h>
#include <ardour/slave.h>
#include <ardour/auditioner.h>

#include "i18n.h"

using namespace ARDOUR;
using namespace SigC;

void
Session::process (jack_nframes_t nframes)
{
	if (synced_to_jack() && waiting_to_start) {
		if ( _engine.transport_state() == AudioEngine::TransportRolling) {
			actually_start_transport ();
		}
	}

	if (non_realtime_work_pending()) {
		if (atomic_read (&butler_should_do_transport_work) == 0) {
			post_transport ();
		} 
	} 
	
	(this->*process_function) (nframes);
}

void
Session::prepare_diskstreams ()
{
	for (DiskStreamList::iterator i = diskstreams.begin(); i != diskstreams.end(); ++i) {
		(*i)->prepare ();
	}
}

int
Session::no_roll (jack_nframes_t nframes, jack_nframes_t offset)
{
	jack_nframes_t end_frame = _transport_frame + nframes;
	int ret = 0;

	if (_click_io) {
		_click_io->silence (nframes, offset);
	}

	/* XXX we're supposed to have the route_lock while doing this.
	   this is really bad ...
	*/

	if (atomic_read (&processing_prohibited)) {
		for (RouteList::iterator i = routes.begin(); i != routes.end(); ++i) {
			(*i)->silence (nframes, offset);
		}
		return 0;
	}

	for (RouteList::iterator i = routes.begin(); i != routes.end(); ++i) {
		
		if ((*i)->hidden()) {
			continue;
		}
		
		if ((*i)->no_roll (nframes, _transport_frame, end_frame, offset, non_realtime_work_pending(), 
				   actively_recording(), get_transport_declick_required())) {
			error << compose(_("Session: error in no roll for %1"), (*i)->name()) << endmsg;
			ret = -1;
			break;
		}
	}

	return ret;
}

int
Session::process_routes (jack_nframes_t nframes, jack_nframes_t offset)
{
	bool record_active = actively_recording();
	int  declick = get_transport_declick_required();
	bool rec_monitors = get_rec_monitors_input();

	if (transport_sub_state & StopPendingCapture) {
		/* force a declick out */
		declick = -1;
	}

	for (RouteList::iterator i = routes.begin(); i != routes.end(); ++i) {

		int ret;

		if ((*i)->hidden()) {
			continue;
		}

		if ((ret = (*i)->roll (nframes, _transport_frame, _transport_frame + nframes, offset, declick, record_active, rec_monitors)) < 0) {

			error << compose(_("Session: error for %1 at frame %2 (%3)"), (*i)->name(), _transport_frame, ret) << endmsg;

			/* we have to do this here. Route::roll() for an AudioTrack will have called DiskStream::process(),
			   and the DS will expect DiskStream::commit() to be called. but we're aborting from that
			   call path, so make sure we release any outstanding locks here before we return failure.
			*/

			for (DiskStreamList::iterator ids = diskstreams.begin(); ids != diskstreams.end(); ++ids) {
				(*ids)->recover ();
			}

			stop_transport ();
			return -1;
		} 
	}

	return 0;
}

void
Session::commit_diskstreams (jack_nframes_t nframes, bool &needs_butler)
{
	int dret;

	for (DiskStreamList::iterator i = diskstreams.begin(); i != diskstreams.end(); ++i) {

		if ((*i)->hidden()) {
			continue;
		}
		
		/* force all diskstreams not handled by a Route to call do their stuff.
		 */

		if ((dret = (*i)->process (nframes, 0, actively_recording(), get_rec_monitors_input())) == 0) {
			if ((*i)->commit (nframes)) {
				needs_butler = true;
			}

		} else if (dret < 0) {
			(*i)->recover();
		}

	}
	
	if (actively_recording()) {
		set_dirty();
	}
}

void
Session::process_with_events (jack_nframes_t nframes)
{
	Event* ev;
	jack_nframes_t this_nframes;
	jack_nframes_t end_frame;
	jack_nframes_t offset;
	bool session_needs_butler = false;
	jack_nframes_t stop_limit;
	long           frames_moved;

	if (auditioner) {
		auditioner->silence (nframes, 0);
	}

	while (pending_events.read (&ev, 1) == 1) {
		merge_event (ev);
	}

	/* if we are not in the middle of a state change,
	   and there are immediate events queued up,
	   process them. 
	*/

	while (!non_realtime_work_pending() && !immediate_events.empty()) {
		Event *ev = immediate_events.front ();
		immediate_events.pop_front ();
		process_event (ev);
	}

	if (!process_can_proceed()) {
		no_roll (nframes, 0);
		return;
	}
		
	if (events.empty() || next_event == events.end()) {
		process_without_events (nframes);
		return;
	}

	end_frame = _transport_frame + nframes;

	{
		TentativeLockMonitor rm (route_lock, __LINE__, __FILE__);
		Event* this_event;
		Events::iterator the_next_one;

		if (!rm.locked() || (post_transport_work & (PostTransportLocate|PostTransportStop))) {
			no_roll (nframes, 0);
			return;
		}
		
		if (_slave) {
			if (!follow_slave ()) {
				no_roll (nframes, 0);
				return;
			}
		} 

		if (_transport_speed == 0) {
			no_roll (nframes, 0);
			return;
		}

		if (actively_recording()) {
			stop_limit = max_frames;
		} else {

			if (Config->get_stop_at_session_end()) {
				stop_limit = current_end_frame();
			} else {
				stop_limit = max_frames;
			}
		}

		if (maybe_stop (stop_limit)) {
			no_roll (nframes, 0);
			return;
		} 

		this_event = *next_event;
		the_next_one = next_event;
		++the_next_one;

		offset = 0;

		while (nframes) {

			if (this_event == 0 || this_event->action_frame > end_frame || this_event->action_frame < _transport_frame) {

				this_nframes = nframes;
				
			} else {
				
				/* compute nframes to next event */
				
				if (this_event->action_frame < end_frame) {
					this_nframes = nframes - (end_frame - this_event->action_frame);
				} else {
					this_nframes = nframes;
				}
			}

			if (this_nframes) {
				
				/* now process frames between now and the first event in this block */
				prepare_diskstreams ();

				if (process_routes (this_nframes, offset)) {
					no_roll (nframes, 0);
					return;
				}
				
				commit_diskstreams (this_nframes, session_needs_butler);

				click (_transport_frame, nframes, offset);
				
				nframes -= this_nframes;
				offset += this_nframes;
				
				frames_moved = (jack_nframes_t) floor (_transport_speed * this_nframes);
			
				if (frames_moved < 0) {
					decrement_transport_position (-frames_moved);
				} else {
					increment_transport_position (frames_moved);
				}

				maybe_stop (stop_limit);
				check_declick_out ();
			}

			/* now handle this event and all others scheduled for the same time */
			
			while (this_event && this_event->action_frame == _transport_frame) {
				process_event (this_event);

				if (the_next_one == events.end()) {
					this_event = 0;
				} else {
					this_event = *the_next_one;
					++the_next_one;
				}
			} 

			/* if an event left our state changing, do the right thing */

			if (non_realtime_work_pending()) {
				no_roll (nframes, offset);
				break;
			}

			/* this is necessary to handle the case of seamless looping */
			/* not sure if it will work in conjuction with varispeed */
			end_frame = _transport_frame + nframes;
			
		}

		set_next_event ();

	} /* implicit release of route lock */

	if (session_needs_butler) {
		summon_butler ();
	} 
	
	if (!_engine.freewheeling() && send_mtc) {
		send_midi_time_code_in_another_thread ();
	}

	return;
}		

bool
Session::follow_slave ()
{
	bool starting;

	if (!_slave->ok()) {
		stop_transport ();
		set_slave_source (None, 0);
		return false;
	}
	
	if (!_slave->locked()) {
		return false;
	}
	
	jack_nframes_t slave_transport_frame = _slave->current_audio_frame ();
	float delta;
	int dir;
	float slave_speed;
	
	if (slave_transport_frame > _transport_frame) {
		delta = slave_transport_frame - _transport_frame;
		dir = 1;
	} else {
		delta = _transport_frame - slave_transport_frame;
		dir = -1;
	}

	slave_speed = _slave->current_speed ();

	if ((starting = _slave->starting())) {
		slave_speed = 0;
	}

	if (Config->get_timecode_source_is_synced()) {
		if (slave_speed != 0) {
			slave_speed = 1.0;
		}
	}

#if 0
	cerr << "delta = " << delta << " dir = " << (dir < 1 ? " ardour ahead" : " master ahead") << " speed = " << slave_speed 
	     << " ts = " << _transport_speed 
	     << " M@"<< slave_transport_frame << " S@" << _transport_frame 
	     << " autoloop ? " << auto_loop 
	     << " starting = " << _slave->starting()
	     << endl;
#endif

	if (delta > _current_frame_rate / 2) { 

		Location* al = _locations.auto_loop_location();

		if (al && auto_loop && (slave_transport_frame < al->start() || slave_transport_frame > al->end())) {
			/* don't jump out of the loop */
			return false;
		}
			
		cerr << "non-linear slave motion - jump to " << slave_transport_frame << endl;
#if 0
		/* aim for 50msecs ahead of the current slave position */
		slave_transport_frame += (jack_nframes_t) floor (0.050 * _current_frame_rate);
		cerr << " adjusted to " << slave_transport_frame << endl;
#endif
		force_locate (slave_transport_frame, (slave_speed != 0 && !starting));
		return false;
	}

	if (Config->get_timecode_source_is_synced()) {
		if (slave_speed == 0) {
			if (!starting) {
				if (_transport_speed != 0.0f) {
					// cerr << "Slave stops transport @ " << _transport_frame << endl;
					if (_slave_type != JACK) {
						last_stop_frame = _transport_frame;
					}
					stop_transport();
				}
			}
		} else if (_transport_speed != 1.0f) {
			// cerr << "Slave starts transport\n";
			start_transport ();
		}
	} else {
		request_transport_speed ((0.9 * _transport_speed) + (0.1 * slave_speed));
	}

	if (starting) {
		return false;
	}

	if (non_realtime_work_pending()) {
		// cerr << "slave said no roll\n";
		return false;
	}

	// cerr << "slave said ROLL, transport speed = " << _transport_speed << endl;
	return true;
}

void
Session::process_without_events (jack_nframes_t nframes)
{
	bool session_needs_butler = false;
	jack_nframes_t stop_limit;
	long frames_moved;
	
	{
		TentativeLockMonitor rm (route_lock, __LINE__, __FILE__);

		if (!rm.locked() || (post_transport_work & (PostTransportLocate|PostTransportStop))) {
			no_roll (nframes, 0);
			return;
		}

		if (_slave) {
			if (!follow_slave ()) {
				no_roll (nframes, 0);
				return;
			}
		} 

		if (_transport_speed == 0) {
			no_roll (nframes, 0);
			return;
		}
		
		if (actively_recording()) {
			stop_limit = max_frames;
		} else {
			if (Config->get_stop_at_session_end()) {
				stop_limit = current_end_frame();
			} else {
				stop_limit = max_frames;
			}
		}
		
		if (maybe_stop (stop_limit)) {
			no_roll (nframes, 0);
			return;
		} 

		prepare_diskstreams ();
	
		frames_moved = (long) floor (_transport_speed * nframes);

		if (process_routes (nframes, 0)) {
			no_roll (nframes, 0);
			return;
		}

		commit_diskstreams (nframes, session_needs_butler);

		click (_transport_frame, nframes, 0);

		if (frames_moved < 0) {
			decrement_transport_position (-frames_moved);
		} else {
			increment_transport_position (frames_moved);
		}

		maybe_stop (stop_limit);
		check_declick_out ();

	} /* implicit release of route lock */

	if (session_needs_butler) {
		summon_butler ();
	} 
	
	if (!_engine.freewheeling() && send_mtc) {
		send_midi_time_code_in_another_thread ();
	}

	return;
}		

void
Session::process_audition (jack_nframes_t nframes)
{
	TentativeLockMonitor rm (route_lock, __LINE__, __FILE__);

	if (rm.locked()) {
		for (RouteList::iterator i = routes.begin(); i != routes.end(); ++i) {
			if (!(*i)->hidden()) {
				(*i)->silence (nframes, 0);
			}
		}
	}
	
	if (auditioner->play_audition (nframes) > 0) {
		summon_butler ();
	} 

	if (!auditioner->active()) {
		process_function = &Session::process_with_events;
	}
}

