/*
    Copyright (C) 2004 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: vst_plugin.cc,v 1.6 2004/05/21 11:36:42 pauld Exp $
*/

#include <algobase.h>
#include <vector>
#include <string>
#include <ctype.h>

#include <cstdlib>
#include <cstdio> // so libraptor doesn't complain
#include <cmath>
#include <dirent.h>
#include <string.h> // for memmove
#include <sys/stat.h>
#include <cerrno>

#include <lrdf.h>
#include <fst.h>

#include <pbd/compose.h>
#include <pbd/error.h>
#include <pbd/pathscanner.h>
#include <pbd/xml++.h>

#include <vst/aeffectx.h>

#include <midi++/manager.h>

#include <ardour/ardour.h>
#include <ardour/session.h>
#include <ardour/audioengine.h>
#include <ardour/vst_plugin.h>

#include <pbd/stl_delete.h>

#include "i18n.h"
#include <locale.h>

using namespace ARDOUR;

VSTPlugin::VSTPlugin (AudioEngine& e, Session& session, FSTHandle* h)
	: Plugin (e, session)
{
	handle = h;
	been_resumed = false;

	if ((_fst = fst_instantiate (handle, Session::vst_callback, this)) == 0) {
		throw failed_constructor();
	}

	_plugin = _fst->plugin;
	_plugin->user = this;

	/* set rate and blocksize */

	_plugin->dispatcher (_plugin, effSetSampleRate, 0, 0, NULL, 
			     (float) session.frame_rate());
	_plugin->dispatcher (_plugin, effSetBlockSize, 0, 
			     session.get_block_size(), NULL, 0.0f);
	
	/* set program to zero */

	_plugin->dispatcher (_plugin, effSetProgram, 0, 0, NULL, 0.0f);
	
	Plugin::setup_midi_controls ();
}

VSTPlugin::VSTPlugin (const VSTPlugin &other)
	: Plugin (other)
{
	handle = other.handle;
	been_resumed = false;

	if ((_fst = fst_instantiate (handle, Session::vst_callback, this)) == 0) {
		throw failed_constructor();
	}
	_plugin = _fst->plugin;

	Plugin::setup_midi_controls ();
}

VSTPlugin::~VSTPlugin ()
{
	deactivate ();
	GoingAway (this); /* EMIT SIGNAL */
	fst_close (_fst);
}

void
VSTPlugin::store_state (PluginState& state)
{
}

void
VSTPlugin::restore_state (PluginState& state)
{
}

float
VSTPlugin::default_value (unsigned long port)
{
	return 0;
}	

void
VSTPlugin::set_parameter (unsigned long which, float val)
{
	_plugin->setParameter (_plugin, which, val);
	ParameterChanged (which, val); /* EMIT SIGNAL */
}

float
VSTPlugin::get_parameter (unsigned long which) const
{
	return _plugin->getParameter (_plugin, which);
	
}

unsigned long
VSTPlugin::nth_parameter (unsigned long n, bool& ok) const
{
	ok = true;
	return n;
}

XMLNode&
VSTPlugin::get_state()
{
	XMLNode *root = new XMLNode (state_node_name());
	return *root;
}

int
VSTPlugin::set_state(const XMLNode& node)
{
	if (node.name() != state_node_name()) {
		error << _("Bad node sent to VSTPlugin::set_state") << endmsg;
		return 0;
	}

	return -1;
}

int
VSTPlugin::get_parameter_descriptor (unsigned long which, ParameterDescriptor& desc) const
{
	VstParameterProperties prop;

	if (_plugin->dispatcher (_plugin, effGetParameterProperties, which, 0, &prop, 0)) {

		if (prop.flags & kVstParameterUsesIntegerMinMax) {
			desc.lower = prop.minInteger;
			desc.upper = prop.maxInteger;
		} else {
			desc.lower = 0;
			desc.upper = 1.0;
		}
		
		if (prop.flags & kVstParameterUsesIntStep) {
			
			desc.step = prop.stepInteger;
			desc.smallstep = prop.stepInteger;
			desc.largestep = prop.stepInteger;
			
		} else if (prop.flags & kVstParameterUsesFloatStep) {
			
			desc.step = prop.stepFloat;
			desc.smallstep = prop.smallStepFloat;
			desc.largestep = prop.largeStepFloat;
			
		} else {
			
			float range = desc.upper - desc.lower;
			
			desc.step = range / 100.0f;
			desc.smallstep = desc.step / 2.0f;
			desc.largestep = desc.step * 10.0f;
		}
		
		desc.toggled = prop.flags & kVstParameterIsSwitch;
		desc.logarithmic = false;
		desc.sr_dependent = false;
		desc.label = prop.label;

	} else {

		/* old style */

		char label[64];
		label[0] = '\0';

		_plugin->dispatcher (_plugin, effGetParamName, which, 0, label, 0);

		desc.label = label;
		desc.integer_step = false;
		desc.lower = 0.0f;
		desc.upper = 1.0f;
		desc.step = 0.01f;
		desc.smallstep = 0.005f;
		desc.largestep = 0.1f;
		desc.toggled = false;
		desc.logarithmic = false;
		desc.sr_dependent = false;
	}

	return 0;
}

list<string>
VSTPlugin::get_presets()
{
	list<string> labels;
	return labels;
}

bool
VSTPlugin::load_preset(const string preset_label)
{
	return true;
}

bool
VSTPlugin::save_preset (string name)
{
	return true;
}

string
VSTPlugin::describe_parameter (unsigned long param)
{
	VstParameterProperties prop;
	_plugin->dispatcher (_plugin, effGetParameterProperties, param, 0, &prop, 0);
	return prop.label;
}

jack_nframes_t
VSTPlugin::latency () const
{
	return 0;
}

set<unsigned long>
VSTPlugin::automatable () const
{
	set<unsigned long> ret;
	return ret;
}

int
VSTPlugin::connect_and_run (vector<Sample*>& bufs, size_t maxbuf, size_t& in_index, size_t& out_index, jack_nframes_t nframes, jack_nframes_t offset)
{
	float *ins[_plugin->numInputs];
	float *outs[_plugin->numOutputs];
	size_t i;

	if( !been_resumed) {
		been_resumed = true;
		_plugin->dispatcher (_plugin, effMainsChanged, 0, 1, NULL, 0.0f);
		cerr << "resumed " << name() << endl;
	}

	for (i = 0; i < (size_t) _plugin->numInputs; ++i) {
		ins[i] = bufs[min(in_index,maxbuf)];
		in_index++;
	}

	for (i = 0; i < (size_t) _plugin->numOutputs; ++i) {
		outs[i] = bufs[min(out_index,maxbuf)];

		/* unbelievably, several VST plugins still rely on Cubase
		   behaviour and do not silence the buffer in processReplacing 
		   when they have output.
		*/
		
		// memset (outs[i], 0, sizeof (Sample) * nframes);
		out_index++;
	}


	/* we already know it can support processReplacing */

	_plugin->processReplacing (_plugin, ins, outs, nframes);
	
	return 0;
}

void
VSTPlugin::deactivate ()
{
	_plugin->dispatcher (_plugin, effMainsChanged, 0, 0, NULL, 0.0f);
}

void
VSTPlugin::activate ()
{
	_plugin->dispatcher (_plugin, effMainsChanged, 0, 1, NULL, 0.0f);
}

const char *
VSTPlugin::name () const
{
	return handle->name;
}

const char *
VSTPlugin::maker () const
{
	return "imadeit";
}

const char *
VSTPlugin::label () const
{
	return handle->name;
}

unsigned long
VSTPlugin::parameter_count() const
{
	return _plugin->numParams;
}

bool
VSTPlugin::has_editor () const
{
	return _plugin->flags & effFlagsHasEditor;
}

void
VSTPlugin::print_parameter (unsigned long param, char *buf, unsigned int len) const
{
	char lab[9];
	char *first_nonws;

	_plugin->dispatcher (_plugin, effGetParamLabel, param, 0, lab, 0);
	_plugin->dispatcher (_plugin, effGetParamDisplay, param, 0, buf, 0);

	if (buf[0] == '\0') {
		return;
	}

	first_nonws = buf;
	while (*first_nonws && isspace (*first_nonws)) {
		first_nonws++;
	}
	if (*first_nonws == '\0') {
		return;
	}
#if 0
	char *last_nonws;
	last_nonws = buf;
	// move to terminating null
	while (*last_nonws) {
		last_nonws++;
	}
	// backup to last real char
	last_nonws--;
	// backup through whitespace
	while (isspace (*last_nonws) && last_nonws != buf) {
		last_nonws--;
	}
	cerr << "after backup, lwn = " << *last_nonws << endl;
	last_nonws++;
	cerr << " ZERO " << *last_nonws << endl;
	*last_nonws = '\0';
#endif

	memmove (buf, first_nonws, strlen (buf) - (first_nonws - buf) + 1);
}
