/*
    Copyright (C) 2004 Paul Barton-Davis

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: mtc.cc,v 1.1 2004/06/11 16:22:18 pauld Exp $
*/

#include <cstdlib>
#include <unistd.h>
#include <string>
#include <iostream>

#include <midi++/types.h>
#include <midi++/parser.h>
#include <midi++/port.h>
#include <midi++/mmc.h>
#include <pbd/transmitter.h>

using namespace std;
using namespace SigC;
using namespace MIDI;

bool
Parser::possible_mtc (byte *sysex_buf, size_t msglen)
{
	if (msglen != 9 || sysex_buf[0] != 0xf0 || sysex_buf[1] != 0x7f || sysex_buf[3] != 0x01 || sysex_buf[4] != 0x01) {
		return false;
	}

        /* full MTC */
	
	_mtc_time[0] = sysex_buf[8]; // frames
	_mtc_time[1] = sysex_buf[7]; // minutes
	_mtc_time[2] = sysex_buf[6]; // seconds
	_mtc_time[3] = (sysex_buf[5] & 0x1f); // hours
	
	_mtc_fps = MTC_FPS ((sysex_buf[5] & 0x60) >> 5); // fps
	
	/* wait for first quarter frame, which could indicate forwards
	   or backwards ...
	*/

	expected_mtc_quarter_frame_code = 0;
	_mtc_running = MTC_Stopped;
	_mtc_locked = false;
	memset (_qtr_mtc_time, 0, sizeof (_qtr_mtc_time));
	
	/* emit signals */

	mtc (*this, &sysex_buf[1], msglen - 1);
	mtc_time (_mtc_time, true);
	mtc_status (MTC_Stopped);

	return true;
}

void
Parser::process_mtc_quarter_frame (byte *msg)
{
	int which_quarter_frame = (msg[1] & 0xf0) >> 4;

	/* Is it an expected frame? If not, stop and wait for the first
	   message of a series. Remember, the first can be frame 7 or frame 0,
	   depending on the direction of the MTC generator ...
	*/

	// cerr << which_quarter_frame << " vs. " << expected_mtc_quarter_frame_code << endl;

	if (which_quarter_frame != expected_mtc_quarter_frame_code) {

		/* go back to waiting for the first frame */

		if (_mtc_running != MTC_Stopped) {

			/* skipped a frame while running, reset */

			expected_mtc_quarter_frame_code = 0;
			memset (_qtr_mtc_time, 0, sizeof (_qtr_mtc_time));

			/* tell listener(s) that we skipped. if they return
			   true, just ignore this in terms of it being an error.
			*/

			if (mtc_skipped ()) {
				return;
			}

			/* otherwise, its a error */
			
			_mtc_running = MTC_Stopped;
			_mtc_locked = false;
			mtc_status (MTC_Stopped);
		}
		return;
	}
	switch (which_quarter_frame) {
	case 0: // frames LS nibble
		_qtr_mtc_time[0] |= msg[1] & 0xf;
		break;
	case 1:  // frames MS nibble
		_qtr_mtc_time[0] |= (msg[1] & 0xf)<<4;
		break;
	case 2: // seconds LS nibble
		_qtr_mtc_time[1] |= msg[1] & 0xf;
		break;
	case 3: // seconds MS nibble
		_qtr_mtc_time[1] |= (msg[1] & 0xf)<<4;
		break;
	case 4: // minutes LS nibble
		_qtr_mtc_time[2] |= msg[1] & 0xf;
		break;
	case 5: // minutes MS nibble
		_qtr_mtc_time[2] |= (msg[1] & 0xf)<<4;
		break;
	case 6: // hours LS nibble
		_qtr_mtc_time[3] |= msg[1] & 0xf;
		break;
	case 7: 

		/* last quarter frame msg has the MS bit of
                   the hour in bit 0, and the SMPTE FPS type
                   in bits 5 and 6 
		*/

		_qtr_mtc_time[3] |= ((msg[1] & 0x1) << 4);
		_mtc_fps = MTC_FPS ((msg[1] & 0x6) >> 1);
		break;

	default:
		/*NOTREACHED*/
		break;

	} 
	
	mtc (*this, &msg[1], msglen - 1);

	if (_mtc_running == MTC_Stopped) {

		switch (which_quarter_frame) {
		case 0:
			_mtc_running = MTC_Forward;
			break;
		case 7:
			_mtc_running = MTC_Backward;
			break;

		default:
			break;
		}
	}

	switch (_mtc_running) {
	case MTC_Forward:
		if (which_quarter_frame == 7) {

			/* we've reached the final of 8 quarter frame messages.
			   store the time, reset the pending time holder,
			   and signal anyone who wants to know the time.
			*/

			memcpy (_mtc_time, _qtr_mtc_time, sizeof (_mtc_time));
			memset (_qtr_mtc_time, 0, sizeof (_qtr_mtc_time));
			if (!_mtc_locked) {
				_mtc_locked = true;
				mtc_status (MTC_Forward);
			}
			mtc_time (_mtc_time, false);
			expected_mtc_quarter_frame_code = 0;

		} else {
			expected_mtc_quarter_frame_code = which_quarter_frame + 1;
		}
		break;
		
	case MTC_Backward:
		if (which_quarter_frame == 0) {
			
			/* we've reached the final of 8 quarter frame messages.
			   store the time, reset the pending time holder,
			   and signal anyone who wants to know the time.
			*/

			memcpy (_mtc_time, _qtr_mtc_time, sizeof (_mtc_time));
			memset (_qtr_mtc_time, 0, sizeof (_qtr_mtc_time));
			if (!_mtc_locked) {
				_mtc_locked = true;
				mtc_status (MTC_Backward);
			}
			mtc_time (_mtc_time, false);
			expected_mtc_quarter_frame_code = 7;

		} else {
			expected_mtc_quarter_frame_code = which_quarter_frame - 1;
		}
		break;

	default:
		break;
	}

}
